
#import "ppc_asm_macros.h"

// PPC constants
#define VRSAVE_SPR 256

// Integer registers (We can use r3-r10 w/o having to push anything on the stack)

// Argument registers
#define ARGS          r3
#define   ZC_ARG      0      // The 2 starting Z and 2 C components, in one vector
#define   XYSTEP_ARG  16     // The 2 xstep and 2 ystep components, in one vector
#define   MAXITER_ARG 32
#define   POSX_ARG    48
#define   POSY_ARG    64
#define   WIDTH_ARG   80
#define   HEIGHT_ARG  84
#define   ROWLEN_ARG  88
#define   IMAGE_ARG   92

#define VRSAVE r4
#define TMP    r5
#define SEL    r6
#define WIDTH  r7
#define HEIGHT r8
#define POSX   r9
#define POSY   r10

// Other integer registers which must be saved and restored
#define XNEXT       r31
#define YNEXT       r30
#define IMAGE       r29
#define ROWLEN      r28
#define SELOFF      r27
#define PIXEL       r26
#define TMP2        r25
#define TMP3        r24
#define ROTATE      r23
#define VCONV       r22
#define PICBASE     r21
#define IREG_LAST   21

// Vector registers
// All registers available for our use.  This is not generally the case, but
// we know that this is true in this program.
#define ZERO     v0
#define ITER     v1
#define RADIUS   v2
#define XR_STEP  v3
#define XI_STEP  v4
#define YR_STEP  v5
#define YI_STEP  v6
#define MAXITER  v7
#define SELMASK  v8
#define TWO      v9
#define MAGS     v10
#define ONEI     v11
#define DONEMASK v12
#define VTMP1    v13
#define VTMP2    v14
#define ZR       v15
#define ZI       v16
#define CR       v17
#define CI       v18
#define ZR_NEXT  v19
#define ZI_NEXT  v20
#define ZR_LEFT  v21
#define ZI_LEFT  v22
#define ZR2      v23
#define ZI2      v24
#define VTMP3    v25


//
// DATA
//

// Selection masks

.data
.align 4
selectionMasks:
    .long 0xffffffff, 0x00000000, 0x00000000, 0x00000000
    .long 0x00000000, 0xffffffff, 0x00000000, 0x00000000
    .long 0x00000000, 0x00000000, 0xffffffff, 0x00000000
    .long 0x00000000, 0x00000000, 0x00000000, 0xffffffff

.align 4
rotate:
     .byte 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
     .byte 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 0, 1, 2, 3
     .byte 8, 9, 10, 11, 12, 13, 14, 15, 0, 1, 2, 3, 4, 5, 6, 7
     .byte 12, 13, 14, 15, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11

.align 4
radius:
    .single 4.0, 4.0, 4.0, 4.0

.align 4
two:
    .single 2.0, 2.0, 2.0, 2.0

.align 4
one:
    .long 1, 1, 1, 1

.data
.align 4
vConvert:
    .long 0, 0, 0, 0




.text
.align 2
.globl _vcComputeJulia4
_vcComputeJulia4:
    // Since we want the code to be usable in dynamically loaded situations
    // we need to reference the data above relative to the instruction counter.
    // via the normal PIC stuff.  This means we always need a stack frame.
    mflr    r0
    stw     r0,8(r1)
    stwu    r1,-64(r1)
    stmw    REG(IREG_LAST),-4*(32-IREG_LAST)(r1)

    bl      L_PicBase
L_PicBase:
    mflr    PICBASE

    // We will be using vector registers.  Let the kernel which vector
    // registers to save.  For now, save them all.
    mfspr   VRSAVE,VRSAVE_SPR
    xor	    TMP,TMP,TMP	// 0
    not     TMP,TMP     // 0xffffffff
    mtspr   VRSAVE_SPR,TMP

    vxor    ZERO,ZERO,ZERO  // Need a zero vector to do plain multiply (w/o adding)

    addis	TMP,PICBASE,ha16(radius - L_PicBase)
    addi 	TMP,TMP,lo16(radius - L_PicBase)
    lvx	RADIUS,0,TMP

    // Load the arguments
    li  TMP,ZC_ARG
    lvx VTMP1,TMP,ARGS  // contains both Zstart and C

    vspltw ZR_NEXT,VTMP1,0
    vmr    ZR_LEFT,ZR_NEXT

    vspltw ZI_NEXT,VTMP1,1
    vmr    ZI_LEFT,ZI_NEXT
    
    vspltw CR,VTMP1,2
    vspltw CI,VTMP1,3

    li  TMP,XYSTEP_ARG
    lvx VTMP1,TMP,ARGS

    vspltw XR_STEP,VTMP1,0
    vspltw XI_STEP,VTMP1,1
    vspltw YR_STEP,VTMP1,2
    vspltw YI_STEP,VTMP1,3

    li   TMP,MAXITER_ARG
    lvx  MAXITER,TMP,ARGS

    addi POSX,ARGS,POSX_ARG
    addi POSY,ARGS,POSY_ARG
    lwz  WIDTH,WIDTH_ARG(ARGS)
    lwz  HEIGHT,HEIGHT_ARG(ARGS)
    lwz  IMAGE,IMAGE_ARG(ARGS)
    lwz  ROWLEN,ROWLEN_ARG(ARGS)

    addis	SEL,PICBASE,ha16(selectionMasks - L_PicBase)
    addi 	SEL,SEL,lo16(selectionMasks - L_PicBase)

    addis	ROTATE,PICBASE,ha16(rotate - L_PicBase)
    addi 	ROTATE,ROTATE,lo16(rotate - L_PicBase)

    addis	VCONV,PICBASE,ha16(vConvert - L_PicBase)
    addi 	VCONV,VCONV,lo16(vConvert - L_PicBase)

    addis	TMP,PICBASE,ha16(two - L_PicBase)
    addi 	TMP,TMP,lo16(two - L_PicBase)
    lvx  TWO,0,TMP

    addis	TMP,PICBASE,ha16(one - L_PicBase)
    addi 	TMP,TMP,lo16(one - L_PicBase)
    lvx  ONEI,0,TMP

    // Phew... have stuff loaded now.
    // Now we need to create the first four input points.  If we inline the code
    // to create them here, we will either have to do a lot of work to make sure
    // we deal with the width correctly (we might have a single pixel width).
    // Instead, we will use the point reloading code below.  To signal to it
    // that our points are empty (and no pixel should be written), we'll start
    // out with the ITER counter set to 0xffffffff for each point.

    vxor   ITER,ITER,ITER     // zero out all of the iteration counters
    vnot   ITER,ITER          // Turn on all bits
    vmr    DONEMASK,ITER      // Mark all points as 'done' (i.e., needing attention)

    li     XNEXT,0            // The next point to consider is the first one
    li     YNEXT,0

    b      L_InsertStart      // Set up the first set of points

L_PointLoop:

L_IterationLoop:

    //
    // Compute an iteration of the Julia equation.  This is the most critical section
    // of the code.  The scheduling here must be very good.
    //
		vmaddfp   VTMP1, ZR, ZI, ZERO
		vmaddfp   ZR2, ZR, ZR, ZERO
		vmaddfp   ZI2, ZI, ZI, ZERO
		vadduwm   ITER,ITER,ONEI
		vmaddfp   ZI, VTMP1, TWO, CI
		vsubfp    ZR, ZR2, ZI2
		vaddfp    MAGS, ZR2, ZI2
		vaddfp    ZR, ZR, CR

		vcmpequw  VTMP2,ITER, MAXITER
		vcmpgtfp  VTMP3,MAGS,RADIUS
		vor       DONEMASK,VTMP2,VTMP3
		vcmpequw. VTMP3,DONEMASK,ZERO

  bt        cr6+lt,L_IterationLoop  // If the lt bit (bit zero) is true, do another loop
    //
    // One or more of our points has finished.  The mask of the finished points is in DONEMASK.
    // Check each of the points in turn and record the results and mask in new points.
    //

L_InsertStart:
    li    TMP,4
    mtctr TMP
    li    SELOFF,0

L_InsertLoop:
    lvx       SELMASK,SELOFF,SEL
    vand      VTMP1,DONEMASK,SELMASK  // check if this point is done.
    vcmpequw. VTMP1,VTMP1,SELMASK  // If the point is done, the resulting mask will be equal to the selmask
                                   // and zero otherwise
    bf        cr6+lt,L_InsertLoopBottom  // Wasn't equal, skip to the next point by jumping to the
                                         // bottom of the loop (which will handle decrementing the loop counter
                                         // and terminating the loop if we are done).


    // This position IS done.  Figure out the index of the finished result.  We can
    // compute this by shifting SELOFF down by 4 bits (dividing by 16).  But we
    // really want the index*sizeof(unsigned long), so we'll only shift by 2.
    srawi     TMP,SELOFF,2

    // We might need to store the result.  If we are just starting, the iteration count
    // for this point will be 0xffffffff.  If this is the case, we should NOT store any
    // iteration count (since this slot hasn't been initialized yet).
    vand      VTMP1,SELMASK,ITER
    vcmpequw. VTMP2,VTMP1,SELMASK
    bt        cr6+lt,L_LoadNewPoint

    // We need to store its results.  First, compute the address to
    // store the iteration counter in.  
    lwzx      TMP2,TMP,POSX      // tmp2 = x
    lwzx      TMP3,TMP,POSY      // tmp3 = y
    mullw     PIXEL,TMP3,WIDTH   // pixel = y * width
    add       PIXEL,PIXEL,TMP2   // pixel = y * width + x

    // We need to extract and store the integer iteration counter.  I don't grok the stvewx
    // instruction totally yet, so we'll rotate the slot we are interested in to the first
    // position, store it and load it back into an integer register.
    lvx       VTMP2,SELOFF,ROTATE
    vperm     VTMP1,ITER,ITER,VTMP2
    stvewx    VTMP1,0,VCONV

    // We now want to write out the counter  to allow the higher level algorithm to do
    // block fills.  If I understood how stvewx really worked, I might be able to just store
    // the value in the right position in the first place, but memory bandwidth really isn't
    // the problem in this computation :)
    lwz       TMP2,0(VCONV)
    slwi      PIXEL,PIXEL,2
    stwx      TMP2,PIXEL,IMAGE

L_LoadNewPoint:

    // We might not have any more work to do.  This is the case if the next point
    // to enqueue is at (0, HEIGHT).
    cmplwi XNEXT,0
    bne    L_StillValid
    cmplw  YNEXT,HEIGHT
    bne    L_StillValid

    // No more pixels left to process.  Zero out any values in this point
    // and zero out the values used to update it during the iteration.
    // This will prevent it from ever being reported as done again, allowing
    // us to continue iterating until all pixels are done.
    vsel   ZR,ZR,ZERO,SELMASK
    vsel   ZI,ZI,ZERO,SELMASK
    vsel   CR,CR,ZERO,SELMASK
    vsel   CI,CI,ZERO,SELMASK
    vsel   ONEI,ONEI,ZERO,SELMASK

    // If ONEI is all zeros, then all the enqueued points have been flushed out
    // and we are *completely* done.
    vcmpequw. VTMP1,ONEI,ZERO
    bt        cr6+lt,L_Done

    // Otherwise, continue looping through any other points that might need inserting
    b         L_InsertLoopBottom

L_StillValid:

    // Now, we need to prepare this new slot with a new point and iteration counter
    vsel   ZR,ZR,ZR_NEXT,SELMASK
    vsel   ZI,ZI,ZI_NEXT,SELMASK
    vaddfp ZR_NEXT,ZR_NEXT,XR_STEP
    vaddfp ZI_NEXT,ZI_NEXT,XI_STEP

    // Zero out the iteration counter for this point
    vsel   ITER,ITER,ZERO,SELMASK

    // Store the indicies of this point
    stwx   XNEXT,TMP,POSX
    stwx   YNEXT,TMP,POSY

    // Update the indicies of the next point to be done
    addi   XNEXT,XNEXT,1
    cmplw  XNEXT,ROWLEN
    blt    L_InsertLoopBottom

    // Ran off the line.  Need to step to the next.  If this was the last scanline, we're
    // really out of pixels to enqueue.  However, there are still potentially some pixels
    // left in the current registers that aren't done.  The issue is how to most easily
    // finish those pixels.  The solution we will use here is to zero out sections of key
    // registers used in performing the calculations.  This will ensure that these points
    // are never considered done and pixels will never be written for these points.  This
    // will allow the other points to complete.  Once all of the slots are zeroed out,
    // we know that we are really done.
    xor    XNEXT,XNEXT,XNEXT
    addi   YNEXT,YNEXT,1

    // Reset the next point position to the left edge
    vsubfp ZR_LEFT,ZR_LEFT,YR_STEP
    vsubfp ZI_LEFT,ZI_LEFT,YI_STEP
    vmr    ZR_NEXT,ZR_LEFT
    vmr    ZI_NEXT,ZI_LEFT

    // Loop through the remaining points
L_InsertLoopBottom:

    addi  SELOFF,SELOFF,16
    bdnz  L_InsertLoop

    b L_PointLoop

L_Done:    
    // Done.  Load our saved non-volatile registers and return
    mtspr   VRSAVE_SPR,VRSAVE
    lmw REG(IREG_LAST),-4*(32-IREG_LAST)(r1)
    la    r1,64(r1)
    lwz   r0,8(r1)
    mtlr  r0
    blr


