// Copyright 2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.
//
// $Header: /Network/Users/Omni/bungi/Library/TimCVS/Tim/Altivec/Julia/OmniJulia/julia.h,v 1.3 2000/10/15 06:45:14 bungi Exp $


#import "clut.h"

typedef enum {
    InSetDistance,
    VeryCloseDistance,
    DeltaCloseDistance,
    NotCloseDistance
} DistanceType;

#define MAX_ITERATIONS 255

typedef struct {
    // inputs
    float              c[4];
    unsigned int       N;
    void              *orbit;

    // outputs
    unsigned int   n;
    DistanceType   mandelbrotDistanceType;
    float          mandelbrotDistanceEstimate; // if not in the set
    unsigned int   cycleLength;                // if in the set
} Julia;

// Compute a bunch of information about a single point.  Much slower
// than the batch computation interfaces below.
void qJulia(Julia *julia);


extern void cComputeJuliaImage4(color_t *image,
                                const float center[2],
                                const float dx[2],
                                const float dy[2],
                                unsigned int pixelsWide,
                                unsigned int pixelsHigh,
                                const float c[2],
                                unsigned int maxIterations,
                                color_t clut[256],
                                int isConnected,
                                int useC);

extern void cComputeJuliaImageDist4(color_t *image,
                                    const float center[2],
                                    const float dx[2],
                                    const float dy[2],
                                    unsigned int pixelsWide,
                                    unsigned int pixelsHigh,
                                    const float c[2],
                                    unsigned int maxIterations,
                                    color_t clut[256],
                                    int isConnected,
                                    int useC);


#if 0

// The routine that does the work of generating an image from the inputs
extern void qComputeJuliaImage(color_t *image,
                               const float center[4],
                               const float dx[4],
                               const float dy[4],
                               unsigned int pixelsWide,
                               unsigned int pixelsHigh,
                               const float c[4],
                               unsigned int maxIterations,
                               color_t clut[256],
                               int useAltivec);

extern void cComputeJuliaImage(color_t *image,
                               const float center[4],
                               const float dx[4],
                               const float dy[4],
                               unsigned int pixelsWide,
                               unsigned int pixelsHigh,
                               const float c[4],
                               unsigned int maxIterations,
                               color_t clut[256]);

extern void cComputeMandelbrotImage(color_t *image,
                                    const float center[4],
                                    const float dx[4],
                                    const float dy[4],
                                    unsigned int pixelsWide,
                                    unsigned int pixelsHigh,
                                    unsigned int maxIterations,
                                    color_t clut[256]);

// A routine identical to the one above, except that the mapping between
// the cartesian coordinates and the pixel coordinates is not assumed
// to be linear.  Instead, a warp matrix is used which contains multipliers
// for the x/y coordinates.
extern void qComputeWarpedJuliaImage(color_t *image,
                                     const WarpPoint *warp,
                                     const float center[4],
                                     const float dx[4],
                                     const float dy[4],
                                     unsigned int pixelsWide,
                                     unsigned int pixelsHigh,
                                     const float *c,
                                     unsigned int maxIterations,
                                     color_t clut[256],
                                     int useAltivec);


extern void qComputeJuliaImage4(color_t *image,
                                const float center[4],
                                const float dx[4],
                                const float dy[4],
                                unsigned int pixelsWide,
                                unsigned int pixelsHigh,
                                const float c[4],
                                unsigned int maxIterations,
                                color_t clut[256]);

#endif
